/***************************************************************************
 *   Copyright (C) 2007  Antony Martin <martin.antony@yahoo.fr>            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU General Public License           *
 *   as published by the Free Software Foundation; either version 2        *
 *   of the License, or (at your option) any later version.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************/

/* Cree le : 11 mars 2007
   derniere modification le 13/01/2008 */

#include <SCE/SCEMinimal.h>

#include <SCE/utils/SCEMath.h>
#include <SCE/utils/SCEString.h>
#include <SCE/utils/SCEMedia.h>
#include <SCE/utils/SCEResources.h>

#include <SCE/core/SCECSupport.h>

#include <SCE/interface/SCETexture.h>


/* ajoute le 14/06/2007 */
static int sce_tex_enabled = 1;     /* booleen, true : textures activees */

/* ajoute le 10/07/2007 */
static SCE_STexture *binded = NULL;

/* ajoute le 07/10/2007 */
static SCE_STexture *used = NULL;
static int used_unit = 0;


/* revise le 06/10/2007 */
int SCE_Init_Texture (void)
{
    return SCE_OK;
}

/* revise le 06/10/2007 */
void SCE_Quit_Texture (void)
{
}


/* ajoute le 06/10/2007 */
void SCE_Texture_Bind (SCE_STexture *tex)
{
    binded = tex;
    /* NOTE: binding de la texture du coeur ? */
}


/* revise le 06/10/2007 */
SCE_STexture* SCE_Texture_Create (SCEenum type, int w, int h)
{
    unsigned int i = 0;
    SCE_CTexData data;
    SCE_STexture *tex = NULL;

    SCE_btstart ();
    tex = SCE_malloc (sizeof *tex);
    if (!tex)
    {
        Logger_LogSrc ();
        SCE_btend ();
        return NULL;
    }

    for (i=0; i<6; i++)
        tex->fb[i] = NULL;

    switch (type)
    {
        case SCE_RENDER_DEPTH:
            type = SCE_DEPTH_BUFFER;
        case SCE_RENDER_COLOR:
            /* ben oui, sinon on ignore le depth */
            if (type == SCE_RENDER_COLOR)
                type = SCE_COLOR_BUFFER;

            tex->fb[0] = SCE_CCreateFramebuffer ();
            if (!tex->fb[0])
            {
                SCE_free (tex);
                Logger_LogSrc ();
                SCE_btend ();
                return NULL;
            }
            SCE_CBindFramebuffer (tex->fb[0]);

            /* s'il s'agit d'une color, on ajoute un depth buffer */
            if (type == SCE_COLOR_BUFFER)
                SCE_CAddRenderBuffer_ (SCE_DEPTH_BUFFER, 0, w, h);

            if (SCE_CCreateRenderTexture_ (type, 0, 0, 0, w, h) < 0)
            {
                SCE_Texture_Delete (tex);
                Logger_LogSrc ();
                SCE_btend ();
                return NULL;
            }
            tex->fb[0]->buffers[0].user = SCE_TRUE; /* TODO: tres laid */
            tex->tex = SCE_CGetRenderTexture_ (type);
            SCE_CBindFramebuffer (NULL);
            
            break;

        case SCE_RENDER_DEPTH_CUBE:
            type = SCE_DEPTH_BUFFER;
        case SCE_RENDER_COLOR_CUBE:
            /* ben oui, sinon on ignore le depth */
            if (type == SCE_RENDER_COLOR_CUBE)
                type = SCE_COLOR_BUFFER;

            tex->tex = SCE_CCreateTexture (SCE_TEX_CUBE);
            if (!tex->tex)
            {
                SCE_Texture_Delete (tex);
                Logger_LogSrc ();
                SCE_btend ();
                return NULL;
            }

            for (i=0; i<6; i++)
            {
                tex->fb[i] = SCE_CCreateFramebuffer ();
                if (!tex->fb[i])
                {
                    SCE_Texture_Delete (tex);
                    Logger_LogSrc ();
                    SCE_btend ();
                    return NULL;
                }

                SCE_CInitTexData (&data);

                /* le target est defini dans AddTextureTexData */
                data.w = w;
                data.h = h;
                data.type = SCE_UNSIGNED_BYTE;

                if (type == SCE_DEPTH_BUFFER)
                {
                    data.fmt = GL_DEPTH_COMPONENT;
                    data.pxf = GL_DEPTH_COMPONENT24;
                }
                else
                    data.fmt = data.pxf = GL_RGBA;

                if (SCE_CAddTextureTexDataDup (tex->tex, SCE_TEX_POSX + i,
                                               &data) < 0)
                {
                    /* la il faut les supprimer les donnees, bicoz elles
                       ne sont pas encore incluses a la texture */
                    SCE_Texture_Delete (tex);
                    Logger_LogSrc ();
                    SCE_btend ();
                    return NULL;
                }
                
                /* s'il s'agit d'une color, on ajoute un depth buffer */
                if (type == SCE_COLOR_BUFFER)
                    SCE_CAddRenderBuffer (tex->fb[i], SCE_DEPTH_BUFFER, 0, w, h);
                
                SCE_CAddRenderTexture (tex->fb[i], type, SCE_TEX_POSX + i,
                                       tex->tex, 0, SCE_FALSE);
            }
            
            break;

        default:
            /* aucun type "render" specifie, c'est une texture normale */
            tex->tex = SCE_CCreateTexture (type);
            if (!tex->tex)
            {
                SCE_free (tex);
                Logger_LogSrc ();
                SCE_btend ();
                return NULL;
            }
    }

    /* assignation de quelques parametres pour les textures cubemap */
    if (type == SCE_RENDER_COLOR_CUBE || type == SCE_TEX_CUBE ||
        type == SCE_RENDER_DEPTH_CUBE)
    {
        SCE_CBindTexture (tex->tex);
        SCE_CSetTextureParam_ (GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
        SCE_CSetTextureParam_ (GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
        SCE_CSetTextureParam_ (GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);
        SCE_CBindTexture (tex->tex);
        /* on remet la precedente texture active */
        SCE_Texture_Use (used, used_unit);
    }

    SCE_btend ();
    return tex;
}


/* ... */
#define SCE_DEFAULTFUNC(action)\
SCE_STexture *back = binded;\
SCE_Texture_Bind (tex);\
action;\
SCE_Texture_Bind (back);

#define SCE_DEFAULTFUNCR(t, action)\
t r;\
SCE_STexture *back = binded;\
SCE_Texture_Bind (tex);\
r = action;\
SCE_Texture_Bind (back);\
return r;

/* ajoute le 07/10/2007 */
void SCE_Texture_Delete (SCE_STexture *tex)
{
    SCE_DEFAULTFUNC (SCE_Texture_Delete_ ())
}
/* revise le 07/10/2007 */
void SCE_Texture_Delete_ (void)
{
    SCE_btstart ();
    if (binded)
    {
        unsigned int i;
        for (i=0; i<6; i++)
            SCE_CDeleteFramebuffer (binded->fb[i]);
        SCE_CDeleteTexture (binded->tex);
        SCE_free (binded);
        binded = NULL;
    }
    SCE_btend ();
}


/* ajoute le 07/10/2007 */
SCE_CTexture* SCE_Texture_GetCTexture (SCE_STexture *tex)
{
    return tex->tex;
}
/* ajoute le 07/10/2007 */
SCE_CTexture* SCE_Texture_GetCTexture_ (void)
{
    return binded->tex;
}


/* ajoute le 07/10/2007 */
int SCE_Texture_Build (SCE_STexture *tex, int use_mipmap)
{
    int hw_mipmap = 0;

    SCE_btstart ();
    SCE_CBindTexture (tex->tex);

    /* par defaut, le moteur tente de faire generer les mipmaps en hardware */
    hw_mipmap = (use_mipmap && SCE_CHasCap (SCE_TEX_HW_GEN_MIPMAP));
    if (SCE_CGetTextureTarget (tex->tex) != SCE_TEX_CUBE)
        hw_mipmap = (SCE_CGetTextureNumMipmaps_ (0) <= 1 && hw_mipmap);

    /* construction */
    if (SCE_CBuildTexture (tex->tex, use_mipmap, hw_mipmap) < 0)
    {
        Logger_LogSrc ();
        SCE_btend ();
        return SCE_ERROR;
    }

    SCE_CBindTexture (NULL);

    SCE_btend ();
    return SCE_OK;
}
/* revise le 07/10/2007 */
int SCE_Texture_Build_ (int use_mipmap)
{
    return SCE_Texture_Build (binded, use_mipmap);
}

/* revise le 07/10/2007 */
int SCE_Texture_Update (SCE_STexture *tex)
{
    SCE_btstart ();
    if (SCE_CUpdateTexture (tex->tex, -1, -1) < 0)
    {
        Logger_LogSrc ();
        SCE_btend ();
        return SCE_ERROR;
    }

    SCE_btend ();
    return SCE_OK;
}
/* ajoute le 07/10/2007 */
int SCE_Texture_Update_ (void)
{
    return SCE_Texture_Update (binded);
}


/* ajoute le 13/01/2008 */
SCE_STexture* SCE_Texture_LoadArg (int type, int w, int h, int d, va_list args)
{
    SCE_STexture *tex = NULL;

    SCE_btstart ();
    tex = SCE_Texture_Create (type, w, h);
    if (!tex)
    {
        Logger_LogSrc ();
        SCE_btend ();
        return NULL;
    }

    /* TODO: !! temporaire !! */
    SCE_CDeleteTexture (tex->tex);
    tex->tex = NULL;
    /* !!!! */

    /**
     * /todo : attention, double creation d'une texture coeur (plus now...)
     */
    tex->tex = SCE_CLoadTextureArg (type, w, h, d, args);
    if (!tex->tex)
    {
        SCE_Texture_Delete (tex);
        Logger_LogSrc ();
        tex = NULL;
    }

    SCE_btend ();
    return tex;
}
/* ajoute le 07/10/2007 */
/* revise le 13/01/2008 */
SCE_STexture* SCE_Texture_Load (int type, int w, int h, int d, ...)
{
    va_list args;
    SCE_STexture *tex = NULL;
    va_start (args, d);
    tex = SCE_Texture_LoadArg (type, w, h, d, args);
    va_end (args);
    return tex;
}

/* ajoute le 14/06/2007 */
void SCE_Texture_Active (int enabled)
{
    sce_tex_enabled = enabled;
}
/* ajoute le 14/06/2007 */
void SCE_Texture_Enable (void)
{
    sce_tex_enabled = SCE_TRUE;
}
/* ajoute le 14/06/2007 */
void SCE_Texture_Disable (void)
{
    sce_tex_enabled = SCE_FALSE;
}

/* revise le 07/10/2007 */
void SCE_Texture_Use (SCE_STexture *tex, unsigned int unit)
{
    if (sce_tex_enabled)
    {
        /* desactivation de la precedente texture */
        SCE_CUseTexture (NULL, unit);

        if (tex)
            SCE_CUseTexture (tex->tex, unit);

        used = tex;
        used_unit = unit;
    }
}


/* ajoute le 04/07/2007 */
/* revise le 07/10/2007 */
void SCE_Texture_RenderTo (SCE_STexture *tex, SCEuint cubeface)
{
    if (tex)
    {
        /* s'il s'agit d'une cubemap, on prend en compte 'cubeface' */
        if (SCE_CGetTextureTarget (tex->tex) == SCE_TEX_CUBE)
            SCE_CUseFramebuffer (tex->fb[cubeface], NULL);
        else
            SCE_CUseFramebuffer (tex->fb[0], NULL);
    }
    else
        SCE_CUseFramebuffer (NULL, NULL);
}
/* ajoute le 07/10/2007 */
void SCE_Texture_RenderTo_ (SCEuint cubeface)
{
    SCE_Texture_RenderTo (binded, cubeface);
}
